
/*
   Copyright (c) 2002 Compaq Computer Corporation
   
   SOFTWARE RELEASE
   
   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software and associated documentation files (the
   "Software"), to deal in the Software without restriction, including
   without limitation the rights to use, copy, modify, merge, publish,
   distribute, sublicense, and/or sell copies of the Software, and to
   permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
   
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   
   - Neither the names of Compaq Research, Compaq Computer Corporation
     nor the names of its contributors may be used to endorse or promote
     products derived from this Software without specific prior written
     permission.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
   IN NO EVENT SHALL COMPAQ COMPUTER CORPORATION BE LIABLE FOR ANY CLAIM,
   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
   THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

package AccordionTreeDrawer;

import java.awt.Color;
import java.io.IOException;
import java.util.*;
import gl4java.*;
import gl4java.GLFunc;

import AccordionDrawer.*;

/**
 * A class representing an edge of a (phylognenetic) tree
 * 
 * A TreeEdge represents a horizontal or vertical line segment used in drawing a tree.
 *
 * <Pre>
 *     a--- firstChild 
 *     |               
 *     |               
 *     ---  .          
 * n   |    .          
 * o---|c   .	       
 * d   ---             
 * e   |               
 *     |               
 *     b--- lastChild  
 * 
 * </Pre>
 *
 * In this figure (node,c), (a,b), (firstChild,a) and (lastChild,b) are line segments
 * Each of them is represented by a TreeEdge
 * (node,c) corresponds to a horizontal TreeEdge associated with TreeNode node
 * (a,b) corresponds to a vertical TreeEdge associated with TreeNode node
 * 
 * @author  Tamara Munzner, Serdar Tasiran, Li Zhang, Yunhong Zhou
 * @version 
 * @see     AccordionDrawer.Tree
 * @see     AccordionDrawer.TreeNode
 * @see     AccordionDrawer.GridCell
 */

public class TreeEdge {

    public TreeNode node;

    /**
    * Whether this TreeEdge is horizontal
    */ 
    public boolean horiz;


    public TreeEdge(TreeNode n, boolean horizontal)
    {
		node = n;
		horiz = horizontal; // from me to parent I'm horizontal
    }
    
//    public TreeNode getNode()
//    {
//    	return node;
//    }

	private GridCell getCell()
	{
		return node.getCell();
	}

	public AccordionTreeDrawer getDrawer()
	{
		return (AccordionTreeDrawer)getCell().drawer;
	}

//	/**
//	 * clean the tree edge
//	 * @see TreeNode.close
//	 *
//	 */
//	public void close(){
//			node=null;
//	}
    
//	protected void finalize() throws Throwable
//	{	 
//		try
//		{
//			close();
//		}
//		finally
//		{
//			super.finalize();
//		}
//	}
 

//    public int getKey() { return node.getKey(); }
//    public String getName() { return node.getName(); } // never used?
//    public GridCell getCell() { return node.getCell();}
//	public void setCell(GridCell c) { node.setCell(c);}
    
    // these will return splitLine indices
    public int getMin() { return node.getMin(); }
    public int getMax() { return node.getMax(); }

	public void setPositions(float[] start, float[] end)//, int direction)
	{
		AccordionTreeDrawer atd = getDrawer();
		int X = AccordionDrawer.X, Y = AccordionDrawer.Y;
		GridCell c = getCell();
		double[] minStuck = {atd.getSplitLine(Y).getMinStuckValue(),
			atd.getSplitLine(Y).getMinStuckValue()};
		double[] rangeSize = {
			atd.getSplitLine(X).getMaxStuckValue() - minStuck[X],
			atd.getSplitLine(Y).getMaxStuckValue() - minStuck[Y]};
		if (horiz)
		{
			start[X] = (float)c.getMin(X); // X is over the range
			end[X] = (float)c.getMax(X);
			start[Y] = end[Y] = (float)node.getMidY();
		}
		else
		{
			start[X] = end[X] = (float)c.getMax(X);
			start[Y] = (float)node.getMinY();
			end[Y] = (float)node.getMaxY();
		}

		// draw in reverse direction
//		if (direction == AccordionDrawer.RIGHT)
//		{
//			start[X] = (float)(rangeSize[X] - start[X] + minStuck[X]);
//			end[X] = (float)(rangeSize[X] - end[X] + minStuck[X]);
//		}
		//TODO: fix drawing in different directions
//		//dawn in up direction
//		else if (direction == AccordionDrawer.UP)
//		{
//			float temp= start[0];
//			start[0]= start[1];
//			start[1] = 1- temp;
//			temp= end[0];
//			end[0]= end[1];
//			end[1] =1- temp;
//		}
//		// draw in down direction
//		else if (direction == AccordionDrawer.DOWN)
//		{
//			float temp= start[1];
//			start[1]= start[0];
//			start[0] =1- temp;
//			temp= end[1];
//			end[1]= end[0];
//			end[0] =1- temp;
//		}
	}
    
    /**
     * Draws this TreeEdge inside the GridCell to which it is
     * attached, and its label if appropriate.
     *
     * Find font size to use for drawing the label, by checking for
     * occlusions. Don't draw it all if it's occluded even at the
     * smallest size.
     * 
     * @author   Tamara Munzner, Li Zhang
     * 
     * @param    col The color to draw this node in
     * @param    plane Z depth at which to draw label
     *
     * @see      AccordionDrawer.GridCell
     */

	private Color setGLColor(AccordionTreeDrawer atd, Color col)
	{
		float rgbcol[] = new float[3];
		Color newrgb;
		if (atd.dimcolors || atd.dimbrite) {
			
			col.getRGBColorComponents(rgbcol);
			float hsbcol[] = Color.RGBtoHSB((int) (rgbcol[0] * 255),
					(int) (rgbcol[1] * 255), (int) (rgbcol[2] * 255),
					(new float[3]));
			float howdim;
			// H stays the same
			// if started out greyscale, don't crank up S
			if (hsbcol[1] > 0f && atd.dimcolors) {
				// depends on size of your subtended subtree screen area
				InteractionBox b = atd.makeBox(node); // box around subtree
				if (null == b)
					System.out.println("interaction box null");
				// howdim base: one pixel 0, fullscreen 1.0
				howdim = (float) (b.getMax(AccordionDrawer.Y) - b
						.getMin(AccordionDrawer.Y))
						/ (float) atd.getWinMax(1); //screen coords
				// small dynamic range:, 0%=.5, 20% and up =1.0
				howdim = .5f + 40 * howdim;
				howdim = (howdim > 1.0) ? 1.0f : howdim;
				hsbcol[1] = (float) howdim;
			}

			if (atd.dimbrite && col == atd.getObjectColor()) {
				// depends on depth in tree
				int treeheight = atd.tree.getHeight();
				// howdim ranges from 0 at root to 1.0 at leaf
				howdim = (node.height - 1.0f) / (treeheight - 1.0f);
				// compress the range
				hsbcol[2] = (float) (0.2 + howdim * .7);
			}
			newrgb = Color.getHSBColor(hsbcol[0], hsbcol[1], hsbcol[2]);
		} else {
			newrgb = col;
		}
		return newrgb;
	}
	
    public void drawInCell(Color col, double plane) {
		//    	System.out.println("Drawing: " + node.getCell());
		int X = horiz ? 0 : 1;
		int Y = horiz ? 1 : 0;
		AccordionTreeDrawer atd = (AccordionTreeDrawer) node.cell.drawer;
		
		boolean isBase = false;
		if (col == null)
		{
			col = atd.objectColor;
			isBase = true;
		}
		
		if (atd.nogeoms)
			return;
		AccordionTreeDrawer.countDrawnFrame++;
		GLFunc gl = (GLFunc) atd.getGraphicsStuff(); // this is not a gl call, just a reference
		Color drawColor = setGLColor(atd, col);

		// draw the line
		int thick = atd.getLineThickness();
		if (col != atd.getObjectColor() && !atd.isDoingFlash())
			thick += 2;
		float[] posStart = new float[2], posEnd = new float[2];
		setPositions(posStart, posEnd);//, atd.drawDirection);
		//	if (horiz)
		//		System.out.println(node + " Set positions: Y:" + posEnd[1]);
		if (atd.takeSnapshot)
		{
			if ((atd.basePass && isBase) || (atd.groupPass && !isBase))
			try
			{
				int cellMinPix[] = {atd.w2s(posStart[X], X), atd.w2s(posStart[Y], Y)};
				int cellMaxPix[] = {atd.w2s(posEnd[X], X), atd.w2s(posEnd[Y], Y)};
				atd.snapShotWriter.write(thick + " setlinewidth newpath " +
						cellMinPix[X] + " " + cellMinPix[Y] + " moveto " +
						cellMaxPix[X] + " " + cellMaxPix[Y] + " lineto " +
						"closepath " +
						"gsave " + 
						drawColor.getRed() / 255f + " " +
						drawColor.getGreen() / 255f + " " +
						drawColor.getBlue() / 255f + " setrgbcolor " +
						"stroke grestore\n");
			}
			catch (IOException ioe)
			{
				System.out.println("Error: IOException while trying to write cell to file: " + atd.snapShotWriter.toString());
			}
		
		}
		else
		{
			atd.setColorGL(drawColor);
			gl.glLineWidth(thick);
			gl.glBegin(GLEnum.GL_LINES);
			gl.glVertex3d(posStart[0], posStart[1], plane);
			gl.glVertex3d(posEnd[0], posEnd[1], plane);
			gl.glEnd();
		}

		if (horiz)
		{ // draw the dot at the end of the horizontal section
			if (atd.takeSnapshot)
			{
				if ((atd.basePass && isBase) || (atd.groupPass && !isBase))
				try
				{
					int cellMaxPix[] = {atd.w2s(posEnd[X], X), atd.w2s(posEnd[Y], Y)};
					atd.snapShotWriter.write("newpath " +
							(cellMaxPix[X]-thick) + " " + (cellMaxPix[Y]-thick) + " moveto " +
							(cellMaxPix[X]-thick) + " " + (cellMaxPix[Y]+thick) + " lineto " +
							(cellMaxPix[X]+thick) + " " + (cellMaxPix[Y]+thick) + " lineto " +
							(cellMaxPix[X]+thick) + " " + (cellMaxPix[Y]-thick) + " lineto " +
							"closepath " +
							"gsave " + 
							drawColor.getRed() / 255f + " " +
							drawColor.getGreen() / 255f + " " +
							drawColor.getBlue() / 255f + " setrgbcolor " +
							"eofill grestore\n");
				}
				catch (IOException ioe)
				{
					System.out.println("Error: IOException while trying to write cell to file: " + atd.snapShotWriter.toString());
				}
			
			}
			else
			{
				gl.glPointSize((float) (thick + 2.0f));
				gl.glBegin(GLEnum.GL_POINTS);
				gl.glVertex3d(posEnd[0], posEnd[1], plane);
				gl.glEnd();
			}
		}

		if (horiz && atd.drawlabels && node.label.length() > 0 && !atd.basePass && !atd.groupPass ) {
			// first, check smallest font, to see if anything at all
			// can fit. if it does, then do binary search to find
			// largest possible box that fits. after we know which box
			// size to use, draw it.

			// this algorithm means drawing order will have a radical
			// effect on how labels look!		
				
				// we should set min to one less than the real bound. new
				// real bound is minFontSize+1 (because we already tested
				// minFontSize), so leave min set to current value of
				// (minFontSize+1)-1 = minFontSize
				
				int min = ((AccordionTreeDrawer)atd).minFontHeight;
				int max = ((AccordionTreeDrawer)atd).maxFontHeight;
				int mid = max;
				int fitsheight = min;
				LabelBox fits = makeLabelBox(min, -1, -1, posStart, posEnd);
//				System.out.println("min font size: " + min);
				if (!intersectLabelBox(fits))
				{
					if (max != min)
					{
						LabelBox maxBox = makeLabelBox(max, -1, -1, posStart, posEnd);
						if (!intersectLabelBox(maxBox))
						{
							fits = maxBox;
							fitsheight = max;
						}
						else while (min+1 < max)
						{
							mid = (int) ((min + max) / 2.0);
//							if (mid == max)
//								break; // stop testing the same value
//							System.out.println("testing font size: " + min + " " + mid + " " + max);
							LabelBox lb = makeLabelBox(mid, -1, -1, posStart, posEnd);
							if (intersectLabelBox(lb))
								max = mid; // too big
							else
							{
								fits = lb; // bigger size
								fitsheight = mid;
								min = mid;
							}
						}
					}
//					if (atd.takeSnapshot)
//						System.out.println(node.name + " using font size: " + fitsheight);
					drawLabelBox(fits, fitsheight, false);
//					else
//						drawLabelBox(fits, fitsheight, false);
				}
		}
	}
    
    /**
     * Draw label of this TreeEdge at maximum size (intended for
     * mouseover highlighting).
     * 
     * @author   Tamara Munzner
     */

    public void drawLabelBig(int x, int y, int fontheight) {
		AccordionTreeDrawer drawer = (AccordionTreeDrawer) getCell().drawer;
		if (!horiz || drawer.nogeoms || node.label.length() < 1)
			return;
		drawLabelBox(drawer.flashLabelBox, fontheight, true);
	}


    /** 
     * Create a LabelBox for the given fontheight
     * @param    x horizontal base location in screen/pixel coordinates
     * @param    y vertical base location in screen/pixel coordinates
     * @author Tamara Munzner
     * @see	 AccordionDrawer.LabelBox
     */

    protected LabelBox makeLabelBox(int fontheight, int x, int y, float[] start, float[] end) {
	    AccordionTreeDrawer d = (AccordionTreeDrawer)getCell().drawer;
	    int X = AccordionDrawer.X, Y = AccordionDrawer.Y;
		int bottomLeft[] = new int[2];
		int topRight[] = new int[2];
		int bottomLeftBackground[] = new int[2];
		int startPix[] = {d.w2s(start[X], X), d.w2s(start[Y], Y)};
		int endPix[] = {d.w2s(end[X], X), d.w2s(end[Y], Y)};
		String name = node.label;
		int namewidth = AccordionDrawer.stringWidth(name, fontheight); // expensive
		int oldnamewidth = 0;
		int labelwidth = namewidth+2;
		int numChars;
		int[] winSize = {d.getWinsize(X), d.getWinsize(Y)};

		int moveover = (node.isLeaf()) ? labelwidth + d.labeloffset[X] : -d.labeloffset[X];
//			if(d.drawDirection != AccordionDrawer.RIGHT)
			{
				topRight[X] = (d.labelposright) ?
				endPix[X]+moveover :
				startPix[X]-d.labeloffset[X];
			}
		// either put label right and above junction
		// or put it left, vertically centered, and right-justified
		if (!d.labelposright)
		{
//			if(d.leftRight == 0 || d.leftRight == 1)
			topRight[Y] = startPix[Y]-d.labeloffset[Y]-fontheight;
//		   if(d.leftRight == 2)
//		   labelTopRight[1] = d.w2s(1-posStart[0],1)-d.labeloffset[1]-fontheight;
//		   if(d.leftRight == 3)
//			labelTopRight[1] = d.w2s(posStart[0],1)-d.labeloffset[1]-fontheight;
		}
		else 				
			topRight[Y] = (node.isLeaf()) ? 
					endPix[Y] - (int)(fontheight/2.0) : 
					endPix[Y]-d.labeloffset[Y] - fontheight;       		
		
		// we pad background label box by one pixel on each side so
		// that text is never right up against the edge. height
		// already has enough padding
		int labelheight = fontheight+2;	
		bottomLeftBackground[X] = topRight[X] - labelwidth;
		bottomLeftBackground[Y] = topRight[Y] + labelheight;
		
	
		if (x>=0 && y>=0 && d.labelpopup) {
			bottomLeftBackground[X] = x-(int)(labelwidth/2.0);
			topRight[X] = bottomLeftBackground[X]+labelwidth;
			bottomLeftBackground[Y] = y-5;
			topRight[Y] = bottomLeftBackground[Y]-labelheight;
		} 

		// if we would start offscreen, nudge over until it's all
		// onscreen. when the origin of the text string is offscreen,
		// opengl won't draw anything (sigh). we could truncate the
		// label text, but it seems more useful to just move the whole
		// thing over. while we're at it, make sure labels are always
		// totally onscreen.

		int labelToEdge = winSize[X] - endPix[X];
		if ((bottomLeftBackground[X]) < 0) {
		   bottomLeftBackground[X] = 0;
		   topRight[X] = bottomLeftBackground[X] + labelwidth;
		} else if (topRight[X] > winSize[X]) {	
			// do nothing - in leaf case, we want to truncate not nudge
			// as per hillis feedback.
		} 
//		System.out.println("top right is less than posend  ############"+labelTopRight[0]);
//		 System.out.println("pos end  ############" + d.w2s(1-posEnd[0], 0));

		if (bottomLeftBackground[Y] > winSize[Y]) {
			bottomLeftBackground[Y] = winSize[Y];
			topRight[Y] = bottomLeftBackground[Y]-labelheight;
		} else if (topRight[Y] < 0) {
			topRight[Y] = 0;
			bottomLeftBackground[Y] = labelheight;
		} 

		// labelbuffer is used for occlusion computations to control
		// label density, but we don't want to draw the bg box that
		// big because it would look silly.

		bottomLeft[Y] = bottomLeftBackground[Y] + d.labelbuffer[Y];
		bottomLeft[X] = bottomLeftBackground[X] - d.labelbuffer[X];
		return new LabelBox( bottomLeft, topRight, bottomLeftBackground, d.getFrameNum(), name);
  }


//TODO: implement drawing label boxes differently, so they work efficiently
//
///**
// * make label box in vertical direction
// * @param fontheight
// * @param x
// * @param y
// * @return
// */
//	private LabelBox makeLabelBoxUp(int fontheight, int x, int y) {
//	AccordionTreeDrawer d = (AccordionTreeDrawer)getCell().drawer;
//	int labelBottomLeft[] = new int[2];
//	int labelTopLeft[] = new int[2];
//	int labelTopRight[] = new int[2];
//	int labelBottomRight[] = new int[2];
//	int labelBottomLeftBG[] = new int[2];
//	int drawVert = 1;
//
////	String name = node.name; // make name here?
//	String name = node.label;
//	int namewidth = AccordionTreeDrawer.bff.stringWidth(name, fontheight);
////	int charwidth = (int)namewidth/name.length();
////	int charheight = fontheight;
//	float ratio = 0.0f;
//	int oldnamewidth = 0;
//	int labelwidth = namewidth+2;
//	int labelheight = fontheight+2;
//	int numChars;
//	int[] winSize = {d.getWinsize(0), d.getWinsize(1)};
//    float temp;  
//    temp = posEnd[0];
//    posEnd[0] = posEnd[1];
//    posEnd[1] = 1- temp;
//    temp = posStart[0];
//    posStart[0]= posStart[1];
//    posStart[1] = 1-temp;
//   
////	int moveover = (node.isLeaf()) ? labelwidth+d.labeloffset[0] 
////		: -d.labeloffset[0];
//	
//	// either put label right and above junction
//	// or put it left, vertically centered, and right-justified
//	if (!d.labelposright)
//	{
////		if(d.leftRight == 0 || d.leftRight == 1)
//		labelTopRight[1] = d.w2s(posStart[1],1)-d.labeloffset[1]-fontheight;
////	   if(d.leftRight == 2)
////	   labelTopRight[1] = d.w2s(1-posStart[0],1)-d.labeloffset[1]-fontheight;
////	   if(d.leftRight == 3)
////		labelTopRight[1] = d.w2s(posStart[0],1)-d.labeloffset[1]-fontheight;
//	}
//	else 		
//	{
//		labelBottomLeft[0] = (node.isLeaf()) ? 
//		d.w2s(posEnd[0],0) - (int)(fontheight/2.0) : 
//		d.w2s(posEnd[0],0)-d.labeloffset[0] - fontheight;
////		labelBottomRight[0] = labelBottomLeft[0] + labelheight;
//	}
//	// we pad background label box by one pixel on each side so
//	// that text is never right up against the edge. height
//	// already has enough padding
//
//	
//		labelBottomLeftBG[0] = d.w2s(posEnd[0], 0);
//		labelBottomLeftBG[1] = d.w2s(posEnd[1], 1);
//		labelTopRight[0] = labelBottomLeftBG[0] + labelheight;
//		labelTopRight[1] = labelBottomLeftBG[1] - labelwidth;
//	
//	if (x>=0 && y>=0 && d.labelpopup) {
//		labelBottomLeftBG[0] = x-(int)(labelwidth/2.0);
//			labelTopRight[0] = labelBottomLeftBG[0]+labelwidth;
//			labelBottomLeftBG[1] = y-5;
//			labelTopRight[1] = labelBottomLeftBG[1]-labelheight;
//	} 
//
//	// if we would start offscreen, nudge over until it's all
//	// onscreen. when the origin of the text string is offscreen,
//	// opengl won't draw anything (sigh). we could truncate the
//	// label text, but it seems more useful to just move the whole
//	// thing over. while we're at it, make sure labels are always
//	// totally onscreen.
//
//	if ((labelBottomLeftBG[0]) < 0) {
////	   labelBottomLeftBG[0] = 0;
////	   labelTopRight[0] = labelBottomLeftBG[0] + labelheight;
//	  
//	} else if (labelTopRight[0] > winSize[0]) {	
//		// do nothing - in leaf case, we want to truncate not nudge
//		// as per hillis feedback.
//	} 
//
//	if (labelBottomLeftBG[1] > winSize[1]) {
////		labelBottomLeftBG[1] = winSize[1];
////		labelTopRight[1] = labelBottomLeftBG[1]-labelwidth;
//	} else if (labelTopRight[1] < 0) {
////		labelTopRight[1] = 0;
////		labelBottomLeftBG[1] = labelTopRight[1]+labelwidth;
//	} 
//	
//	
//	// labelbuffer is used for occlusion computations to control
//	// label density, but we don't want to draw the bg box that
//	// big because it would look silly.
//
//	labelBottomLeft[1] = labelBottomLeftBG[1] + d.labelbuffer[1];
//	labelBottomLeft[0] = labelBottomLeftBG[0] - d.labelbuffer[0];
//
//    if(d.getDrawAngle() < 25.0)
//    {
//	float w = labelBottomLeft[1]*(float)Math.cos(Math.toRadians(d.getDrawAngle()));
//	float h = labelBottomLeft[1]*(float)Math.sin(Math.toRadians(d.getDrawAngle()));
//	labelTopRight[0] = labelTopRight[0] + (int)w;
//	labelTopRight[1] = labelBottomLeft[1] - (int)h; 	
////	labelTopRight[1] =labelBottomLeft[1] - labelBottomLeft[1]*(int)h;
////	labelTopRight[0] = labelBottomLeft[0] + labelBottomLeft[1]*(int)w;
//	if(labelTopRight[1]<0) labelTopRight[1] =0;
//    int cwidth = AccordionTreeDrawer.bff.charWidth(name,fontheight,0);
//	//labelBottomRight[0] = labelBottomLeft[0] + labelheight;
//	labelBottomRight[0] = labelBottomLeft[0] + cwidth;
//	labelBottomRight[1] = labelBottomLeft[1] ;
////	float newTR = labelTopRight[1]* java.lang.Math.sin(java.lang.Math.toRadians(30.0));
//	
//	labelTopLeft[1] = labelTopRight[1];
//	labelTopLeft[0] = labelTopRight[0]-labelheight;
//	lb = new  LabelBox(labelBottomLeft,  labelBottomRight, labelTopLeft, labelTopRight ,labelBottomLeftBG, d.getFrameNum(),  name, labelheight, d.getDrawAngle(), drawVert);
//	    }
//    else
//	lb = new LabelBox( labelBottomLeft, labelTopRight, labelBottomLeftBG, d.getFrameNum(), name);
//
// 	return lb;
//	}
//
//
//	/**
//	 * make label box in vertical direction
//	 * @param fontheight
//	 * @param x
//	 * @param y
//	 * @return
//	 */
//		private LabelBox makeLabelBoxDown(int fontheight, int x, int y) {
//
//    AccordionTreeDrawer d = (AccordionTreeDrawer)getCell().drawer;
//	int labelBottomLeft[] = new int[2];
//	int labelTopLeft[] = new int[2];
//	int labelTopRight[] = new int[2];
//	int labelBottomRight[] = new int[2];
//	int labelBottomLeftBG[] = new int[2];
//	int drawVert = 1;
//
////	String name = node.name; // make name here?
//	String name = node.label;
//	int namewidth = AccordionTreeDrawer.bff.stringWidth(name, fontheight);
////	int charwidth = (int)namewidth/name.length();
////	int charheight = fontheight;
//	float ratio = 0.0f;
//	int oldnamewidth = 0;
//	int labelwidth = namewidth+2;
//	int labelheight = fontheight+2;
//	int numChars;
//	int[] winSize = {d.getWinsize(0), d.getWinsize(1)};
//	float temp;  
//	temp = posEnd[1];
//	posEnd[1] = posEnd[0];
//	posEnd[0] = 1- temp;
//	temp = posStart[1];
//	posStart[1]= posStart[0];
//	posStart[0] = 1-temp;
////	int moveover = (node.isLeaf()) ? labelwidth+d.labeloffset[0] 
////		: -d.labeloffset[0];
//	
//	// either put label right and above junction
//	// or put it left, vertically centered, and right-justified
//	if (!d.labelposright)
//	{
////		if(d.leftRight == 0 || d.leftRight == 1)
//		labelTopRight[1] = d.w2s(posStart[1],1)-d.labeloffset[1]-fontheight;
////	   if(d.leftRight == 2)
////	   labelTopRight[1] = d.w2s(1-posStart[0],1)-d.labeloffset[1]-fontheight;
////	   if(d.leftRight == 3)
////		labelTopRight[1] = d.w2s(posStart[0],1)-d.labeloffset[1]-fontheight;
//	}
//	else 		
//	{
//		labelBottomLeft[0] = (node.isLeaf()) ? 
//		d.w2s(posEnd[0],0) - (int)(fontheight/2.0) : 
//		d.w2s(posEnd[0],0)-d.labeloffset[0] - fontheight;
////		labelBottomRight[0] = labelBottomLeft[0] + labelheight;
//	}
//	// we pad background label box by one pixel on each side so
//	// that text is never right up against the edge. height
//	// already has enough padding
//
//	
//		labelBottomLeftBG[0] = d.w2s(posEnd[0], 0);
//		labelBottomLeftBG[1] = d.w2s(posEnd[1], 1);
//		labelTopRight[0] = labelBottomLeftBG[0] + labelheight;
//		labelTopRight[1] = labelBottomLeftBG[1] - labelwidth;
//	
//	if (x>=0 && y>=0 && d.labelpopup) {
//		labelBottomLeftBG[0] = x-(int)(labelwidth/2.0);
//			labelTopRight[0] = labelBottomLeftBG[0]+labelwidth;
//			labelBottomLeftBG[1] = y-5;
//			labelTopRight[1] = labelBottomLeftBG[1]-labelheight;
//	} 
//
//	// if we would start offscreen, nudge over until it's all
//	// onscreen. when the origin of the text string is offscreen,
//	// opengl won't draw anything (sigh). we could truncate the
//	// label text, but it seems more useful to just move the whole
//	// thing over. while we're at it, make sure labels are always
//	// totally onscreen.
//
//	if ((labelBottomLeftBG[0]) < 0) {
////	   labelBottomLeftBG[0] = 0;
////	   labelTopRight[0] = labelBottomLeftBG[0] + labelheight;
//	  
//	} else if (labelTopRight[0] > winSize[0]) {	
//		// do nothing - in leaf case, we want to truncate not nudge
//		// as per hillis feedback.
//	} 
//
//	if (labelBottomLeftBG[1] > winSize[1]) {
////		labelBottomLeftBG[1] = winSize[1];
////		labelTopRight[1] = labelBottomLeftBG[1]-labelwidth;
//	} else if (labelTopRight[1] < 0) {
////		labelTopRight[1] = 0;
////		labelBottomLeftBG[1] = labelTopRight[1]+labelwidth;
//	} 
//	
//	
//	// labelbuffer is used for occlusion computations to control
//	// label density, but we don't want to draw the bg box that
//	// big because it would look silly.
//
//	labelBottomLeft[1] = labelBottomLeftBG[1] + d.labelbuffer[1];
//	labelBottomLeft[0] = labelBottomLeftBG[0] - d.labelbuffer[0];
//
//	if(d.getDrawAngle() < 25.0)
//	{
//	float w = labelBottomLeft[1]*(float)Math.cos(Math.toRadians(d.getDrawAngle()));
//	float h = labelBottomLeft[1]*(float)Math.sin(Math.toRadians(d.getDrawAngle()));
//	labelTopRight[0] = labelTopRight[0] + (int)w;
//	labelTopRight[1] = labelBottomLeft[1] - (int)h; 	
////	labelTopRight[1] =labelBottomLeft[1] - labelBottomLeft[1]*(int)h;
////	labelTopRight[0] = labelBottomLeft[0] + labelBottomLeft[1]*(int)w;
//	if(labelTopRight[1]<0) labelTopRight[1] =0;
//	int cwidth = AccordionTreeDrawer.bff.charWidth(name,fontheight,0);
//	//labelBottomRight[0] = labelBottomLeft[0] + labelheight;
//	labelBottomRight[0] = labelBottomLeft[0] + cwidth;
//	labelBottomRight[1] = labelBottomLeft[1] ;
////	float newTR = labelTopRight[1]* java.lang.Math.sin(java.lang.Math.toRadians(30.0));
//	
//	labelTopLeft[1] = labelTopRight[1];
//	labelTopLeft[0] = labelTopRight[0]-labelheight;
////	labelBottomLeft[1] = winSize[1] - labelBottomLeft[1];
////	labelBottomRight[1] = winSize[1] - labelBottomRight[1];
////	labelTopLeft[1] = winSize[1] - labelTopLeft[1];
////	labelTopRight[1] = winSize[1]  - labelTopRight[1];
//
//	lb = new LabelBox(labelBottomLeft,  labelBottomRight, labelTopLeft, labelTopRight ,labelBottomLeftBG, d.getFrameNum(),  name, labelheight, d.getDrawAngle(), drawVert);
//		}
//	else
//	lb = new LabelBox( labelBottomLeft, labelTopRight, labelBottomLeftBG, d.getFrameNum(), name);
//
//	return lb;
//
//				}


    /** 
     * Occlusion check of LabelBox against array of all drawn labels. 
     * {@link #drawnArray AccordionDrawer.drawnArray} is sorted in y -
     * the skinny direction, to avoid work. We do a binary search to
     * check for occlusions, so it's log in the number of pixels (with
     * a very small constant).
     * 
     * @author  Tamara Munzner
     * @see AccordionDrawer.LabelBox
     */
    private boolean intersectLabelBox(LabelBox lb) {
	AccordionTreeDrawer d = (AccordionTreeDrawer)getCell().drawer;

	// the list is sorted by y value. use binary search to see
	// if it will fit instead of checking against every item. 
	// find an item that has y overlap. check down from there
	// until no overlap, then check up until no overlap. stop
	// as soon as there's overlap, of course.
	boolean intersects = false;
	ArrayList drawnLabels = d.getDrawnLabels();
	int listlength = drawnLabels.size();
	//int high = listlength-1;
	int high=(listlength==1)?listlength: listlength-1;
	int low = 0;
	int current = -1;
	int prevcurrent = -2;
	int foundIndex = -1;
	LabelBox found;
	//while ((low < high) && (current != prevcurrent) && (!intersects)) {
	while ((low < high) && (!intersects) && (current<=high) && (foundIndex < 0))
	{
	    prevcurrent = current;
	    current = (int)((high+low)/2.0);
	    if (current == prevcurrent && low+1 == high) {
			low=high;
			current = high;
	    }
		if(current == listlength) current = current-1;
	    LabelBox prevlab = (LabelBox) drawnLabels.get(current);
	    //System.out.println("lb bl " + lb.bl[1] + " p tr " + prevlab.tr[1] + " lb tr " + lb.tr[1] + " p bl " + prevlab.bl[1]);
	    if (lb.compareBtoT(prevlab, low, high)) high = current;
	    else if (prevlab.compareBtoT(lb, low, high)) low = current;
	    else { // check intersections
		found = prevlab;
		foundIndex = current;
		//System.out.println("found: " + foundIndex);
		// we're doing a bit more work than necessary by
		// checking against prevlab from last iteration,
		// but it's the easiest way to ensure corner cases
		// work
		while (!intersects && prevlab.bottomBiggerThanTop(lb) && current >= 0) { 
		    // check down from found
		    // i.e. index in sorted list gets smaller, so
		    // y values get smaller also
		    prevlab = (LabelBox) drawnLabels.get(current);
		    intersects = lb.intersectBoxes(prevlab);
		    current--;
		    //System.out.println("cur-- now " + current);
		}
		//current = foundIndex+1;
		current = foundIndex;
		while (!intersects && lb.bottomBiggerThanTop(prevlab) && current <= listlength-1) { 
		    // check up from found
		    prevlab = (LabelBox) drawnLabels.get(current);
		    intersects = lb.intersectBoxes(prevlab);
		    current++;
		    //System.out.println("cur++ now " + current);
		}
	    }
	    //System.out.println("low " + low + " high " + high + " cur " + current + " prevcur " + prevcurrent + " intersects " + intersects);
 	}
	return intersects;
    }

    /** Draw a LabelBox. 
     * @author	 Tamara Munzner, Li Zhang
     * @param	 lb LabelBox to use
     * @param    fontheight Size of font to use, in points/pixels.
     * @param	 drawBig whether to draw maximum size ignoring occlusions
     */
    private void drawLabelBox(LabelBox lb, int fontheight, boolean drawBig) 
    {
    	int X = AccordionDrawer.X, Y = AccordionDrawer.Y;
		AccordionTreeDrawer d = (AccordionTreeDrawer)getCell().drawer;
		GLFunc gl = (GLFunc) d.getGraphicsStuff();
		ArrayList drawnLabels = d.getDrawnLabels();
		if (!drawBig)
		{
	    	drawnLabels.add(lb);
	    	Collections.sort(drawnLabels);
		}
		
		int[] pos = { lb.pos(0), lb.pos(1) };
		int[] topRightPos = { lb.topRightPos(0), lb.topRightPos(1) };

		// highlighted node drawing == drawBig

		double labelplane = d.getLabelplane();
		if (d.labeldrawback || drawBig) {
			float thecol[] = new float[3];
			if (drawBig)
				d.getLabelBackHiColor().getRGBColorComponents(thecol);
			else
				d.getLabelBackColor().getRGBColorComponents(thecol);
			if (!drawBig && d.labeltransp) {
				gl.glColor4f(thecol[0], thecol[1], thecol[2], .5f);
			} else {
				gl.glColor3f(thecol[0], thecol[1], thecol[2]);
			}
//			if (d.drawDirection != AccordionDrawer.RIGHT) {
				gl.glBegin(GLEnum.GL_POLYGON);
				gl.glVertex3d(d.s2w(pos[0], 0), d.s2w(pos[1], 1), labelplane);
				gl.glVertex3d(d.s2w(topRightPos[0], 0), d.s2w(pos[1], 1),
						labelplane);
				gl.glVertex3d(d.s2w(topRightPos[0], 0),
						d.s2w(topRightPos[1], 1), labelplane);
				gl.glVertex3d(d.s2w(pos[0], 0), d.s2w(topRightPos[1], 1),
						labelplane);
				gl.glEnd();
//			}
//			if (d.drawDirection == AccordionDrawer.RIGHT) {
//				gl.glBegin(GLEnum.GL_POLYGON);
//				gl.glVertex3d(d.s2w(pos[0], 0), d.s2w(pos[1], 1), labelplane);
//				gl.glVertex3d(d.s2w(topRightPos[0], 0), d.s2w(pos[1], 1),
//						labelplane);
//				gl.glVertex3d(d.s2w(topRightPos[0], 0), d
//						.s2w(topRightPos[1], 1), labelplane);
//				gl.glVertex3d(d.s2w(pos[0], 0), d.s2w(topRightPos[1], 1),
//						labelplane);
//				gl.glEnd();
//			}

		}
		int descent = AccordionDrawer.getDescent(fontheight);
		String name = lb.getName();
		double b[] = { d.s2w(pos[X] + 1, X), d.s2w(pos[Y] - descent - 1, Y)};
		if (drawBig) // no shadow on highlight
		{
			d.setColorGL(d.getLabelHiColor());
			d.drawText(b[X], b[Y], name, fontheight, d.getLabelColor(), labelplane, null);
		}
		else if (!d.labeldrawback) // shadow, not big, or background box
		{
			d.drawText(b[X], b[Y], name, fontheight, d.getLabelColor(), labelplane, d.getLabelBackColor());
		}
		else // not big, background box so no shadow needed
			d.drawText(b[X], b[Y], name, fontheight, d.getLabelColor(), labelplane, null);
	}
	
// TODO: is there a better way of drawing labels in different directions?
//	
//	
///**
// * draw label box in vertical direction
// * @param d
// * @param gl
// * @param lb
// * @param fontheight
// * @param drawBig
// */
//private void drawLabelBoxUp(AccordionTreeDrawer d,GLFunc gl, LabelBox lb, int fontheight, boolean drawBig)
//	{
//		int[] pos = {lb.pos(0), lb.pos(1)};       
//		int[] topRightPos = {lb.topRightPos(0), lb.topRightPos(1)};
//		//String name = lb.getName();
//		String name = getName();
//		int cwidth;
//        float[][] p = new float[name.length()][];
////		for(int i=0; i<name.length(); i++)
////		{
////		 cwidth= AccordionTreeDrawer.bff.charWidth(name,fontheight,i);
////		 xInterval[i] = cwidth;         
////		 float t =  cwidth*(float)java.lang.Math.tan(java.lang.Math.toRadians(d.getDrawAngle()));
////		 yInterval[i] = (int)t;
////		}
////		d.xInterval = xInterval;
////		d.yInterval = yInterval;
//		if (drawBig) { // save off the pixels we're about to overwrite
//
//			d.drawnlabelsize[0] =fontheight+6;
//			d.drawnlabelsize[1] =fontheight;
//			d.drawnlabelpos[0] = d.s2w(pos[0]-3,0);
//			d.drawnlabelpos[1] = d.s2w(pos[1],1);
//	        int x=pos[0]-3;
//	        int y=pos[1];
//	        for(int i = 0; i<name.length(); i++)
//	        {
//			d.pixels = new float[d.drawnlabelsize[0]*3*d.drawnlabelsize[1]*3];
//            p[i]=d.pixels;
//			gl.glReadBuffer(GLEnum.GL_FRONT);
//			gl.glReadPixels(x,d.getWinsize(1)-y,d.drawnlabelsize[0],d.drawnlabelsize[1], GLEnum.GL_RGB, GLEnum.GL_FLOAT, p[i]);
////			x+=xInterval[i];
////			y-=yInterval[i];
//	        }
//	        d.pal = p;
//	
//		}
//
//		
//		// padding of one pixel on right and bottom, and make sure
//		// we leave enough vertical room for character parts that
//		// descend below font baseline.
//		double labelplane = d.getLabelplane();
//		if (d.labeldrawback || drawBig) {
//			float thecol[] = new float[3];
//			if (drawBig) d.getLabelBackHiColor().getRGBColorComponents(thecol);
//			else d.getLabelBackColor().getRGBColorComponents(thecol);
//			if (!drawBig && d.labeltransp) {
//			gl.glColor4f(thecol[0], thecol[1], thecol[2], .5f);
//		
//		
//		
//			} else {
//			gl.glColor3f(thecol[0], thecol[1], thecol[2]);
//			}
//
//            int x=0;
//            int y=0;
//            int c;
//            for(int i=0; i<lb.getName().length(); i++)
//            {
//			
//			c=10;
//			gl.glDisable(GLEnum.GL_DEPTH _TEST);
//			gl.glBegin(GLEnum.GL_POLYGON);
//			gl.glVertex3d(d.s2w(pos[0]+x-3,0), d.s2w(pos[1]-y,1), labelplane);
//			gl.glVertex3d(d.s2w(pos[0]+x+c+3,0), d.s2w(pos[1]-y,1), labelplane);
//			gl.glVertex3d(d.s2w(pos[0]+x+c+3,0), d.s2w(pos[1]-y-fontheight,1), labelplane);
//			gl.glVertex3d(d.s2w(pos[0]+x-3,0), d.s2w(pos[1]-y-fontheight,1), labelplane);
//			gl.glEnd();
//			gl.glEnable(GLEnum.GL_DEPTH _TEST);
////			x+=xInterval[i];
////			y+=yInterval[i];
//            }
//
//		}
//		int descent = AccordionTreeDrawer.bff.getDescent(fontheight);
//		if (drawBig) 
//			d.setColorGL(d.getLabelHiColor());
//		else {
//			if (!d.labeldrawback) {
//			// first draw 4X (nw/ne/sw/se) in outline color 
//			d.setColorGL(d.getLabelBackColor());
//			int by = pos[1]-descent-1;
//			int bx = pos[0] ;
//	
//		}
//			d.setColorGL(d.getLabelColor());
//		}
//	
//			int x=0;
//			int y =descent+1 ;	
//			for(int i=0; i<name.length(); i++)
//			{				            
//				gl.glRasterPos3d((float)d.s2w(pos[0] +x,0), (float)d.s2w(pos[1]-y,1),   
//				(float)labelplane);
//				AccordionTreeDrawer.bff.drawString(gl, name.substring(i,i+1), fontheight);
//				
////                x+=xInterval[i];
////                y+=yInterval[i];
//			}
//	}
//	
//// draw in down direction
//	
//private void drawLabelBoxDown(AccordionTreeDrawer d,GLFunc gl, LabelBox lb, int fontheight, boolean drawBig)
//	{
//		int[] pos = {lb.pos(0), lb.pos(1)};
//		int[] topRightPos = {lb.topRightPos(0), lb.topRightPos(1)};
//		String name = lb.getName();
//		float[][] p = new float[name.length()][]; 		
//		int cwidth;
//
////		for(int i=0; i<name.length(); i++)
////		{
////			 cwidth= AccordionTreeDrawer.bff.charWidth(name,fontheight,i);
////			 xInterval[i] = cwidth;         
////			 float t =  cwidth*(float)Math.tan(java.lang.Math.toRadians(d.getDrawAngle()));
////		 	yInterval[i] = (int)t;
////		}
////		
////		d.xInterval = xInterval;
////		d.yInterval = yInterval;
//		if (drawBig) { // save off the pixels we're about to overwrite
//   
//			d.drawnlabelsize[0] =fontheight+6;
//			d.drawnlabelsize[1] =lb.size(0)+2;
//			d.drawnlabelpos[0] = d.s2w(pos[0]-3,0);
//			d.drawnlabelpos[1] = d.s2w(pos[1]+fontheight+4,1);
//			int x=pos[0]-3;
//			int y=pos[1]+fontheight+4;
//			for(int i = 0; i<lb.getName().length(); i++)
//			{
//				d.pixels = new float[d.drawnlabelsize[0]*3*d.drawnlabelsize[1]*3];
//
//    	    	p[i] = d.pixels;
//				gl.glReadBuffer(GLEnum.GL_FRONT);
//
//				gl.glReadPixels(x,d.getWinsize(1)-y,d.drawnlabelsize[0],d.drawnlabelsize[1], GLEnum.GL_RGB, GLEnum.GL_FLOAT, p[i]);
////				x+=xInterval[i];
////				y+=yInterval[i];
//				if(y>d.getWinsize(1)) y=d.getWinsize(1)-1;
//			}
//             d.pal = p;
//	
//		}
//
//		
//		// padding of one pixel on right and bottom, and make sure
//		// we leave enough vertical room for character parts that
//		// descend below font baseline.
//		double labelplane = d.getLabelplane();
//		if (d.labeldrawback || drawBig) {
//			float thecol[] = new float[3];
//			if (drawBig) d.getLabelBackHiColor().getRGBColorComponents(thecol);
//			else d.getLabelBackColor().getRGBColorComponents(thecol);
//			if (!drawBig && d.labeltransp) {
//			gl.glColor4f(thecol[0], thecol[1], thecol[2], .5f);
//			
//			
//			} else {
//			gl.glColor3f(thecol[0], thecol[1], thecol[2]);
//			}
//			int x=0;
//			int y=0;
//			int c;
//			for(int i=0; i<lb.getName().length(); i++)
//			{
//				//c=AccordionTreeDrawer.bff.charWidth(lb.getName(),fontheight,i);
//				c=10;
//				gl.glDisable(GLEnum.GL_DEPTH _TEST);
//				gl.glBegin(GLEnum.GL_POLYGON);
//				gl.glVertex3d(d.s2w(pos[0]+x-3,0), d.s2w(pos[1]+y,1), labelplane);
//				gl.glVertex3d(d.s2w(pos[0]+x+c+3,0), d.s2w(pos[1]+y,1), labelplane);
//				gl.glVertex3d(d.s2w(pos[0]+x+c+3,0), d.s2w(pos[1]+y+fontheight,1), labelplane);
//				gl.glVertex3d(d.s2w(pos[0]+x-3,0), d.s2w(pos[1]+y+fontheight,1), labelplane);
//				gl.glEnd();
//				gl.glEnable(GLEnum.GL_DEPTH _TEST);
////				x+=xInterval[i];
////				y+=yInterval[i];
//			}
//	
//		
//		}
//		int descent = AccordionTreeDrawer.bff.getDescent(fontheight);
//		if (drawBig) 
//			d.setColorGL(d.getLabelHiColor());
//		else {
//			if (!d.labeldrawback) {
//			// first draw 4X (nw/ne/sw/se) in outline color 
//			d.setColorGL(d.getLabelBackColor());
//		int by = pos[1]+descent+10;
//			int bx = pos[0] ;
//	
//
//		}
//			d.setColorGL(d.getLabelColor());
//		}
//	
//			int x=0;
//			int y =descent+1 ;	
//			for(int i=0; i<name.length(); i++)
//			{				            
//				gl.glRasterPos3d((float)d.s2w(pos[0] +x,0), (float)d.s2w(pos[1]+y+10,1),   
//				(float)labelplane);
//				AccordionTreeDrawer.bff.drawString(gl, name.substring(i,i+1), fontheight);
////				x+=xInterval[i];
////				y+=yInterval[i];
//			
//			}
//	}

// TODO: this is actually only required by treenode
//	/**
//	 * Determines whether the window coordinates (x,y) fall on this TreeEdge
//	 * There is a "snap" factor, determined by the pickFuzz field of
//	 * the enclosing cell.
//	 *
//	 * @author   Tamara Munzner
//	 * @see      AccordionDrawer.GridCell
//	 * @see      AccordionDrawer.AccordionTreeDrawer
//	 */			
//	// convert everything to window coords (0,1)
//	// x and y are initially pixels
//	public boolean pick(int x, int y)
//	{
//		GridCell c = getCell();
//		AccordionTreeDrawer d = (AccordionTreeDrawer)c.drawer;
//		if (horiz)
//		{ // horizontal line, check in Y direction
//			double pickFuzzAbsolute = d.s2w(d.pickFuzz, AccordionDrawer.Y);
//			double yAbsolute = d.s2w(y, AccordionDrawer.Y);
//			return yAbsolute > c.getMin(AccordionDrawer.Y) - pickFuzzAbsolute &&
//				yAbsolute < c.getMax(AccordionDrawer.Y) + pickFuzzAbsolute;
//		}
//		else
//		{
//			double pickFuzzAbsolute = d.s2w(d.pickFuzz, AccordionDrawer.X);
//			double xAbsolute = d.s2w(x, AccordionDrawer.X);
//			return xAbsolute > c.getMin(AccordionDrawer.X) - pickFuzzAbsolute &&
//				xAbsolute < c.getMax(AccordionDrawer.X) + pickFuzzAbsolute;
//
//		}
//	}


	public void close()
	{
	}
	
	public String toString()
	{
		GridCell c = getCell();
		if (horiz)
		{
			return "X" + c.getMin(AccordionDrawer.X) + "->" + c.getMax(AccordionDrawer.X) + "@Y" + node.getMidY();
		}
		else
		{
			return "Y" + c.getMin(AccordionDrawer.Y) + "->" + c.getMax(AccordionDrawer.Y) + "@X" + c.getMax(AccordionDrawer.X);
		}
	}
	

};

