
/*
   Copyright (c) 2002 Compaq Computer Corporation
   
   SOFTWARE RELEASE
   
   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software and associated documentation files (the
   "Software"), to deal in the Software without restriction, including
   without limitation the rights to use, copy, modify, merge, publish,
   distribute, sublicense, and/or sell copies of the Software, and to
   permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
   
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   
   - Neither the names of Compaq Research, Compaq Computer Corporation
     nor the names of its contributors may be used to endorse or promote
     products derived from this Software without specific prior written
     permission.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
   IN NO EVENT SHALL COMPAQ COMPUTER CORPORATION BE LIABLE FOR ANY CLAIM,
   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
   THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

package AccordionTreeDrawer;

import java.awt.*;
import java.util.*;

import AccordionDrawer.*;

/**
 * A class representing a node of a (phylognenetic) tree.
 * The tree that this node belongs to is of type Tree.
 * Nodes have fields that store a pre- and post-ordering. 
 *
 * A TreeNode has a list of children, a unique key, a leftmostleaf 
 * and a rightmost leaf
 *
 * @author  Tamara Munzner, Li Zhang, Yunhong Zhou
 * @version 
 * @see     Tree
 * @see     GridCell
 */

public class TreeNode extends Object implements CellGeom, Comparable {

    ArrayList children; // eventually turn this into an array (need to change parser)
    TreeEdge edges[] = new TreeEdge[2];
//    LabelBox lb;

    /** key is unique for nodes in one tree */
    public int key; 

    /** The GridCell that this node is attached to  */ 
    public GridCell cell; 

    /** 
     * The index of the smallest descendant node (Node in subtree rooted
     * at this node).
     */
    private int min; 

    /** The index of the largest descendant node */ 
    private int max; 
    
    private int fontSize;
    Double bcnScore;
//    /**
//     * The offset of the point with respect to the cell. We only have
//     * this for the row offset as we assume that the vertical edges
//     * are all aligned.  When computing the Y coordinate of a node, we
//     * add nodeOffsetR to the pointOffset[1], a fixed parameter set by
//     * AccordionDrawer.
//     */
	int computedFrame; // store frame midYPosition was last calculated (needed to place parents)
    double midYPosition;


    /** @return The index of the smallest descendant node */ 
    public int getMin() { return min; }

    /** @return The index of the smallest descendant node */ 
    public int getMax() { return max; }

    public int getKey() { return key;}
    public String getName() { return name;}

    public boolean pick(int x, int y) { return false; }

    /**
     * Draws this node inside cell, the GridCell that it is attached
     * to, if cell is drawn large enough Prints the label of the node
     * if the GridCell is drawn large enough
     * @author   Tamara Munzner, Serdar Tasiran, Li Zhang, Yunhong Zhou
     * 
     * @param    col The color to draw this node in
     * @param    plane The plane to draw this node in
     *
     * @see      AccordionDrawer.GridCell
     */
    
    public void drawInCell(Color col, double plane) {
	for (int xy = 0; xy < 2; xy++) {
	    TreeEdge e = edges[xy];
	    if (null != e) e.drawInCell(col, plane);
	}
    }
    
    public void drawInCell(ArrayList col, double plane)
    {
    	if (cell.getDrawnFrame() >= cell.drawer.getFrameNum())
    		return; // already drawn in this frame
    	cell.setDrawnFrame(cell.drawer.getFrameNum());
    	if (isLeaf())
    		((AccordionTreeDrawer)cell.drawer).leafDrawCount++;
    	else
			((AccordionTreeDrawer)cell.drawer).internalDrawCount++;
    	Color c = null;
    	if (col != null && col.size() > 0)
    	{
    		c = (Color)col.get(0);
    		plane = cell.drawer.hiliteplane - .004*c.getAlpha();
    	}
    	drawInCell(c, plane);
    }

    // cell max[X] boundary set to min[X] boundary over all children
    public void setCell(GridCell c) { cell = c;}
    public GridCell getCell() { return cell;}

    /** Add edge in either X/horizontal (xy==0) direction, or
     * Y/vertical (xy==1)direction. Leaf nodes have no vertical edge. 
     * @author Tamara Munzner */
    public void addEdge(TreeEdge edge, int xy) {
	edges[xy] = edge;
    }
    public TreeEdge getEdge(int xy) { return edges[xy];}

    /** implement Comparable interface - sort on key field */
    public int compareTo(Object o) {
	if (key == ((TreeNode)o).key) return 0;
	else if (key < ((TreeNode)o).key) return -1;
	else return 1;
    }

    public TreeNode parent;

    /** 
     * node name with default "". 
     * Most internal nodes have no name and all leaf nodes have a name 
     */  
    String name = ""; // the long form in fully qualified names
    public String label = ""; // always short form

    int height;

    /** weight is actually edge length for the edge immediately above the node */
    public float weight = 0.0f;

    public TreeNode leftmostLeaf, rightmostLeaf;
    public int numberLeaves;
    public TreeNode preorderNext = null;

    public TreeNode posorderNext = null;

// this can be extracted from split cell position
//    /** leaf index */
//    public int lindex; 

// change next pointers to list of leaves with leaf indices corresponding to
// position in split cells
//    /** next leaf added by Li */
//    public TreeNode next=null;


    /**
     * Children list initial capacity 2 as in most case binary.
     */
    // used in 2 places:
    // - create the root when creating the tree
    // - the parser uses this to create nodes attached to the root
    public TreeNode() {

	children = new ArrayList(2); 
	max = -1;
	min = Integer.MAX_VALUE;
	bcnScore = new Double(0.0);
	//	nodeOffsetR = 0;
    }

	/**
	 * clean the node itself and the tree edge attached with it.
	 * @see TreeEdge.close
	 * @see Tree.close
	 *
	 */   
	   public void close(){
		   children.clear();
		   children=null;
		   name=null;
		   cell = null;
		   if (edges != null)
		   		for(int i=0; i<edges.length; i++)
		   			if(edges[i] !=null)
		   				edges[i].close();
		   edges=null;		
			parent=null;	
			label = null;
			leftmostLeaf=null;
			rightmostLeaf=null;	
			preorderNext = null;
			posorderNext = null;		
			bcnScore=null;
//		   this = null;
		  }
   
		  protected void finalize() throws Throwable {
		 
					   try {
						   close();
					   }
					   finally {
						   super.finalize();     
//						   System.out.println("finally clean treeNodes");
					   }
		   }

    /**
     * Allocate a new TreeNode object with node name and weight.   
     * @param s The name of the node given by a string
     * @param w The weight of the edge immediately above the node
     */
    // used only in copying trees
    public TreeNode(String s, float w){
	children = new ArrayList(2);
	name = new String(s);
	weight = w;
	max = -1;
	min = Integer.MAX_VALUE;
	// nodeOffsetR = 0;
    }

	// not used
    public TreeNode(String s){
	children = new ArrayList(2);
	name = new String(s);
	max = -1;
	min = Integer.MAX_VALUE;
	// nodeOffsetR = 0;
    }

    public void setName(String s){
	name = new String(s);
    }

    public int numberChildren() {
	return children.size();
    }

    public TreeNode getChild(int i){
	return (TreeNode) children.get(i);
    }

    public TreeNode getLeftmostLeaf() {
	
	TreeNode n = this;
	while(!n.isLeaf()) n= n.firstChild();
	return n;
    }

    public TreeNode getRightmostLeaf() {

	TreeNode n = this;
	while(!n.isLeaf()) n = n.lastChild();
	return n;
    }

    public boolean isLeaf() {
	return children.isEmpty();
    }

    public boolean isRoot() {
	return (null == parent);
    }


    /** @return the total number of leaves decedant to this node */
    public int getLeafCount() {	
	if(this == null)
	    return 0;
	else if(isLeaf())
	    return 1;
	else{
	    int sum = 0;
	    for(int i=0; i<numberChildren(); i++)
		sum += getChild(i).getLeafCount();
	    return sum;
	}
	
    }

    public boolean equals(TreeNode n){
	return (name.equals(n.name));
    }

    public void addChild(TreeNode n) {
	children.add(n);
	n.parent = this;
    }

    public TreeNode parent() {
	return parent;
    }

    public void setWeight(double w){
	weight = (float) w;
    }

    public float getWeight() { return weight; }
    
    public TreeNode firstChild() {
	return (TreeNode) children.get(0);
    }

    public TreeNode lastChild() {
	return (TreeNode) children.get(children.size()-1);
    }

    /** 
     * @return <code>true</code> if this is an ancestor of <code>that</code>
     * @param that Another TreeNode object 
     */
    public boolean isAncestorOf(TreeNode that) {
	if(leftmostLeaf.getLindex() > that.leftmostLeaf.getLindex() ||
	  	rightmostLeaf.getLindex() < that.rightmostLeaf.getLindex()) return false;
	return true;
	
    }

    /**
     * Compute the lowest common ancestor between this node and <code>that</that>.
     * The two nodes must belong to the same tree.
     * @author   Li Zhang
     * 
     * @param    that A TreeNode in the Tree that this TreeNode belongs to
     * @return   the lowest common ancestor between this node and "that"
     */
    
    public TreeNode lca(TreeNode that) {
	if(that.isAncestorOf(this)) return that;
	TreeNode current = this;
	while(current != null) {
	    if(current.isAncestorOf(that)) return current;
	    current = current.parent;
	}
	return null;
    }


    /**
     * Compute the lowest common ancestor between this leaf and "that"
     * The two nodes must belong to the same tree and must be leaves
     * @author   Li Zhang
     * 
     * @param    that A TreeNode in the Tree that this TreeNode belongs to
     * @return   the lowest common ancestor between this leaf and <code>that</code>, null if one of the nodes is not a leaf 
     */
    public TreeNode leafLca(TreeNode that) {


	if(!isLeaf()) return null;
	if(!that.isLeaf()) return null;
	TreeNode current = this;
	int a = that.getLindex();
	if(getLindex() > that.getLindex()) {
	    current = that;
	    a = getLindex();
	}
	for(;current!=null;current = current.parent) {
	    if(current.rightmostLeaf.getLindex()>=a)
		return current;
	}
	return null;
    }

    public void print() {
	if(name != null)
	    System.out.print("node name: " + name + "\t");
	else 
	    System.out.print("node name null,\t");
	System.out.println("key: " + key);
    }

    public void printSubtree(){

	print();
	
	for(int i=0; i< children.size(); i++)
	    getChild(i).printSubtree();
    }
       
    public void setSubtreeExtremeLeaves() {
	if(isLeaf()) {
	    leftmostLeaf = this;
	    rightmostLeaf = this;
	    return;
	}

	int i;
	for(i=0; i<children.size();i++) {
	    getChild(i).setSubtreeExtremeLeaves();
	}
	leftmostLeaf = firstChild().leftmostLeaf;
	rightmostLeaf = lastChild().rightmostLeaf;
    }

    
    void setSubtreeMinMax() {
	if(isLeaf()) {
	    min = key;
	    max = key;
	    return;
	}
	for(int i=0; i<children.size();i++) {
	    getChild(i).setSubtreeMinMax();
	    int childMin = getChild(i).min;
	    min = (childMin < min) ? childMin : min;
	    int childMax = getChild(i).max;
	    max = (childMax > max) ? childMax : max;
	}
    }

    /** set numberLeaves field for each subnode */
    int setSubtreeNumberLeaves()
    {
    	numberLeaves = 0;
    	if(isLeaf())
    		numberLeaves = 1; 
    	else
    		for(int i=0; i<children.size(); i++)
    			numberLeaves += getChild(i).setSubtreeNumberLeaves();
    	return numberLeaves;
    }
    
    public String toString()
    {
    	String edge[] = {edges[0]!=null?edges[0].toString():"X", edges[1]!=null?edges[1].toString():"Y"};
    	return name + "(" + key + " @ " + height + ")"+ edge[0] + "&" + edge[1];
    }
    
	public void setFontSize(int fs){ fontSize=fs;}
	   public int getFontSize(){ return fontSize;}
    
	   public boolean isInteriorNode() {
		   return children.size()>0?true:false;
	   }

    public void setBcnScore(float n){
    	bcnScore =new Double(n);
    }
    
    public Double getBcnScore(){
    	return bcnScore;
    }
    
    // where the horizontal line is drawn
    // this CAN'T be recursive, that would be too expensive
    public double getMidY()
    {
    	if (computedFrame >= cell.drawer.getFrameNum())
    		return midYPosition;
    	if (isLeaf())
    	{
    		midYPosition = getSize(AccordionDrawer.Y) * 0.5 + cell.getMin(AccordionDrawer.Y);
    	}
		else if (children.size() == 1)
		{
			midYPosition = ((TreeNode)children.get(0)).getMidY();
		}
		else
		{
    		// halfway between the offsets of end children:
    		TreeNode child0 = (TreeNode)children.get(0);
    		TreeNode childN = (TreeNode)children.get(children.size() - 1);
			midYPosition = (child0.cell.getMax(AccordionDrawer.Y) + childN.cell.getMin(AccordionDrawer.Y))/2;
		}
		computedFrame = cell.drawer.getFrameNum();
		return midYPosition;
    }
    
    // where the vertical line stops
    public double getMinY()
    {
    	if (isLeaf()) return -1.0; // no vertical line for leaf
    	return ((TreeNode)children.get(0)).getMidY();
    }

	// where the vertical line stops
	public double getMaxY()
	{
		if (isLeaf()) return -1.0; // no vertical line for leaf
		return ((TreeNode)children.get(children.size()-1)).getMidY();
	}
	
	public double getSize(int xy)
	{
		return cell.getMax(xy) - cell.getMin(xy);
	}
    
	// called on the root node only
    public TreeNode pickNode(double x, double y, double xFuzz, double yFuzz)
    {
    	if (isNodePicked(x, y, xFuzz, yFuzz))
    		return this; // picked this cell
    	if (!isLeaf() && // root == leaf
    			x > cell.getMax(AccordionDrawer.X) && // x too shallow, go deeper
    			xyInRange(y, 0.0, AccordionDrawer.Y)) // y is in range, go deeper, no fuzz
    		return pickDescend(x, y, xFuzz, yFuzz);
    	return null;
    }
    
    public boolean isNodePicked(double x, double y, double xFuzz, double yFuzz)
    {
		double min[] = {cell.getMin(AccordionDrawer.X), cell.getMin(AccordionDrawer.Y)};
		double max[] = {cell.getMax(AccordionDrawer.X), cell.getMax(AccordionDrawer.Y)};
    	double midY = getMidY(); // where to draw the horizontal line
    	return (x > max[AccordionDrawer.X] - xFuzz && x < max[AccordionDrawer.X] + xFuzz &&
    			y > min[AccordionDrawer.Y] - yFuzz && y < max[AccordionDrawer.Y] + yFuzz) ||
	    		(x > min[AccordionDrawer.X] - xFuzz && x < max[AccordionDrawer.X] + xFuzz &&
	    			y > midY - yFuzz && y < midY + yFuzz);
    }
    
    private boolean xyInRange(double value, double fuzz, int xy)
    {
    	return cell.getMin(xy) - fuzz <= value && cell.getMax(xy) + fuzz >= value;
    }

    // removed recursion
    // needs to be stack based for picking to left/right of large subtrees
    public TreeNode pickDescend(double x, double y, double xFuzz, double yFuzz)
    {
    	Stack pickingStack = new Stack();
    	pickingStack.push(this);
    	while (pickingStack.size() > 0)
    	{
    		TreeNode currRoot = (TreeNode)pickingStack.pop(); // next root to check
    		if (currRoot.isNodePicked(x, y, xFuzz, yFuzz))
    			return currRoot;
    		if (currRoot.isLeaf() || // unpicked leaf
    				!currRoot.xyInRange(y, yFuzz, AccordionDrawer.Y) || // bad Y value
    				currRoot.cell.getMin(AccordionDrawer.X) > x) // bad X value
    		{
    			continue;
    		}
    		// some child must have y in range
    		int minChild = 0;
    		int maxChild = currRoot.children.size() - 1;
    		int midChild = (minChild + maxChild + 1) / 2;
    		TreeNode currChild = (TreeNode)(currRoot.children.get(midChild));
    		while (minChild != maxChild && // converged to currChild, currChild is descended or picked
    				!currChild.xyInRange(y, 0.0, AccordionDrawer.Y)) // binary search in Y, no fuzz
    		{ // find appropriate child
    			if (currChild.cell.getMin(AccordionDrawer.Y) > y) // curr child is too big, max = mid
    				maxChild = midChild;
    			else
    				minChild = midChild;
    			if (minChild+1==maxChild && midChild==minChild)
    				midChild = maxChild;
    			else
    				midChild = (minChild + maxChild) / 2;
    			currChild = (TreeNode)(currRoot.children.get(midChild));
    		}
    		if (currChild.isNodePicked(x, y, xFuzz, yFuzz))
    		{
    			return currChild; // found something
    		}
    		if (midChild > 0)
    			pickingStack.push(currRoot.children.get(midChild-1));
    		if (midChild < currRoot.children.size()-1)
    			pickingStack.push(currRoot.children.get(midChild+1));
    		pickingStack.push(currChild); // the next root
    	}
    	return null;
    }

	// assuming this is a leaf, return the leaf index (the split position in Y direction, actually)
	public int getLindex()
	{
		if (cell == null)
			System.out.println("Error when finding Lindex for " + this);
		return cell.getMaxLine(AccordionDrawer.Y);
	}
	
	// don't draw self, just call this on parents after calling drawDescend
	// ascend drawing until the root is drawn or the node drew in the frame already
	public void drawAscend(int frameNum, float plane)
	{
		if (cell.getDrawnFrame() < frameNum)
		{
			drawInCell(cell.drawer.getColorsForCellGeom(this), plane);
			if (!isRoot())
				parent.drawAscend(frameNum, plane);
			cell.setDrawnFrame(frameNum);
		}
	}
	
	// don't draw self, just call this
	public void drawDescend(int frameNum, float plane, int min, int max)
	{
		drawInCell(cell.drawer.getColorsForCellGeom(this), plane);
		TreeNode currNode = this;
		// binary search
		int minChild = 0;
		int maxChild = currNode.children.size() - 1;
		int midChild = (minChild + maxChild) / 2;
		while (!currNode.isLeaf())
		{
			TreeNode midNode = currNode.getChild(midChild);
			if (midNode.key <= max && // mid node is either some tree above max or is an ancestor of max 
					midNode.rightmostLeaf.key >= min) // midnode has some leaf bigger than min 
			{
				midNode.drawInCell(cell.drawer.getColorsForCellGeom(midNode), plane);
				currNode = midNode;
				minChild = 0;
				maxChild = currNode.children.size() - 1;
			}
			else if (midNode.key <= max) // mid node too small, make it bigger
				minChild = midChild+1;
			else
				maxChild = midChild-1;
			midChild = (minChild + maxChild) / 2;
		}		
	}

	/* (non-Javadoc)
	 * @see AccordionDrawer.CellGeom#drawInCell(double)
	 */
	public void drawInCell(double plane, boolean doFlash) {
		// TODO Auto-generated method stub
		
	}
    
}
